import argparse # pip3 install argparse
import hashlib
import os
import pickle
import time

import torch
import torch.nn.functional as F
import numpy as np
import matplotlib
import torchvision
from scipy.spatial.distance import cdist
from torch.hub import download_url_to_file
from tqdm import tqdm # pip install tqdm
import random
import sklearn.manifold

import matplotlib.pyplot as plt
plt.rcParams["figure.figsize"] = (15, 10)
plt.style.use('dark_background')

import torch.utils.data
import scipy.misc
import scipy.ndimage
import sklearn.decomposition
from sklearn.model_selection import train_test_split


parser = argparse.ArgumentParser(add_help=False)
parser.add_argument('-run_path', default='', type=str)

parser.add_argument('-num_epochs', default=100, type=int)
parser.add_argument('-batch_size', default=512, type=int)
parser.add_argument('-classes_count', default=20, type=int)
parser.add_argument('-samples_per_class', default=0, type=int)

parser.add_argument('-learning_rate', default=1e-4, type=float)

parser.add_argument('-z_size', default=32, type=int)
parser.add_argument('-margin', default=0.2, type=float)

parser.add_argument(
    '-is_debug', default=False, type=lambda x: (str(x).lower() == 'true'))

args, _ = parser.parse_known_args()

RUN_PATH = args.run_path
BATCH_SIZE = args.batch_size
EPOCHS = args.num_epochs
LEARNING_RATE = args.learning_rate
Z_SIZE = args.z_size
MARGIN = args.margin
TRAIN_TEST_SPLIT = 0.8

DEVICE = 'cuda'
MAX_LEN = args.samples_per_class
MAX_CLASSES = args.classes_count # 0 = include all
IS_DEBUG = args.is_debug

if len(RUN_PATH):
    RUN_PATH = f'{int(time.time())}_{RUN_PATH}'
    os.makedirs(RUN_PATH)

if not torch.cuda.is_available() or IS_DEBUG:
    MAX_LEN = 100 # per class for debugging
    MAX_CLASSES = 10 # reduce number of classes for debugging
    DEVICE = 'cpu'
    BATCH_SIZE = 12


class DatasetEMNIST(torch.utils.data.Dataset):
    def __init__(self):
        super().__init__() # 62 classes
        self.data = torchvision.datasets.EMNIST(
            root='../data',
            split='byclass',
            train=(MAX_LEN == 0),
            download=True
        )
        class_to_idx = self.data.class_to_idx
        idx_to_class = dict((value, key) for key, value in class_to_idx.items())
        self.labels = [idx_to_class[idx] for idx in range(len(idx_to_class))]

    def __len__(self):
        return len(self.data)

    def __getitem__(self, idx):
        # list tuple np.array torch.FloatTensor
        pil_x, y_idx = self.data[idx]
        np_x = np.array(pil_x)
        np_x = np.expand_dims(np_x, axis=0) # (1, W, H)

        x = torch.FloatTensor(np_x)
        y = torch.LongTensor([y_idx])
        return x, y


dataset_full = DatasetEMNIST()

labels_train, labels_test = train_test_split(
    dataset_full.labels[-MAX_CLASSES:],
    train_size=TRAIN_TEST_SPLIT,
    shuffle=True,
    random_state=0
)
labels_count = dict((key, 0) for key in dataset_full.labels)

print(
    f'labels_train: {labels_train} '
    f'labels_test: {labels_test} '
)

idx_train = []
idx_test = []
str_args = [str(it) for it in [MAX_LEN, MAX_CLASSES]]
hash_args = hashlib.md5((''.join(str_args)).encode()).hexdigest()
path_cache = f'../data/{hash_args}.pkl'
if os.path.exists(path_cache):
    print('loading from cache')
    with open(path_cache, 'rb') as fp:
        idx_train, idx_test = pickle.load(fp)

else:
    for idx, (x, y_idx) in tqdm(enumerate(dataset_full), 'splitting dataset', total=len(dataset_full)):
        y_idx = y_idx.item()
        label = dataset_full.labels[y_idx]
        if MAX_LEN > 0: # for debugging
            if labels_count[label] >= MAX_LEN:
                if all(it >= MAX_LEN for it in labels_count.values()):
                    break
                continue
        labels_count[label] += 1
        if label in labels_train:
            idx_train.append(idx)
        elif label in labels_test:
            idx_test.append(idx)

    with open(path_cache, 'wb') as fp:
        pickle.dump((idx_train, idx_test), fp)

dataset_train = torch.utils.data.Subset(dataset_full, idx_train)
dataset_test = torch.utils.data.Subset(dataset_full, idx_test)

data_loader_train = torch.utils.data.DataLoader(
    dataset=dataset_train,
    batch_size=BATCH_SIZE,
    shuffle=True,
    drop_last=(len(dataset_train) % BATCH_SIZE < 12)
)

data_loader_test = torch.utils.data.DataLoader(
    dataset=dataset_test,
    batch_size=BATCH_SIZE,
    shuffle=True,
    drop_last=(len(dataset_test) % BATCH_SIZE < 12)
)


class Model(torch.nn.Module):
    def __init__(self):
        super().__init__()
        self.encoder = torchvision.models.resnet18( # 3, W, H
            pretrained=True
        )
        self.encoder.fc = torch.nn.Linear(
            in_features=self.encoder.fc.in_features,
            out_features=Z_SIZE
        )

    def forward(self, x):
        z = self.encoder.forward(x.expand(x.size(0), 3, 28, 28))
        y_prim = 0
        norm = torch.norm(z.detach(), p=1, dim=-1, keepdim=True)
        z = z / norm
        return y_prim, z

model = Model()
optimizer = torch.optim.Adam(model.parameters(), lr=LEARNING_RATE)

model = model.to(DEVICE)

metrics = {}
for stage in ['train', 'test']:
    for metric in ['loss', 'z', 'y', 'acc', 'dist_pos', 'dist_neg', 'loss_neg', 'loss_pos']:
        metrics[f'{stage}_{metric}'] = []

for epoch in range(1, 100):

    metrics_epoch = {key: [] for key in metrics.keys()}

    for data_loader in [data_loader_train, data_loader_test]: # just for classification example
        stage = 'train'
        if data_loader == data_loader_test:
            stage = 'test'

        for x, y_idx in tqdm(data_loader, desc=stage):
            x = x.to(DEVICE)
            y_idx = y_idx.squeeze().to(DEVICE)
            _, z = model.forward(x)

            dist_pos = []
            dist_neg = []
            losses_neg = []
            losses_pos = []


            for i, y_idx_i in enumerate(y_idx):
                z_not_used = z[i+1:]
                y_not_used = y_idx[i+1:]
                if len(z_not_used) == 0:
                    break

                # D_all = 1.0 - F.cosine_similarity(
                #     z[i].expand(z_not_used.size()),
                #     z_not_used
                # )


                D_all = F.pairwise_distance(
                    z[i].expand(z_not_used.size()),
                    z_not_used,
                    p=1
                )
                D_neg_all = D_all[y_not_used != y_idx_i]
                if len(D_neg_all) > 0:
                    # if I want some constraints I could filter samples here
                    j = torch.argmin(D_neg_all) # could also just select all samples within margin or hard samples
                    D_n = D_neg_all[j]
                    dist_neg += D_neg_all.cpu().data.numpy().tolist()

                    losses_neg.append(
                        torch.maximum(MARGIN - D_n, torch.zeros_like(D_n))
                    )

                D_pos_all = D_all[y_not_used == y_idx_i]
                if len(D_pos_all) > 0:
                    j = torch.argmax(D_pos_all) # opportunity to avoid collapse
                    D_p = D_pos_all[j]

                    losses_pos.append(
                        D_p
                    )

                    D_pos_all = D_pos_all[D_pos_all != 0]
                    dist_pos += D_pos_all.cpu().data.numpy().tolist()


            losses_pos_each = torch.mean(torch.stack(losses_pos))
            losses_neg_each = torch.mean(torch.stack(losses_neg))
            loss = torch.mean(
                torch.mean(torch.stack(losses_pos), dim=-1) +
                torch.mean(torch.stack(losses_neg), dim=-1)
            )

            metrics_epoch[f'{stage}_dist_pos'].append(np.mean(dist_pos))
            metrics_epoch[f'{stage}_dist_neg'].append(np.mean(dist_neg))
            metrics_epoch[f'{stage}_loss'].append(loss.cpu().item())
            metrics_epoch[f'{stage}_loss_pos'].append(losses_pos_each.cpu().item())
            metrics_epoch[f'{stage}_loss_neg'].append(losses_neg_each.cpu().item())

            if data_loader == data_loader_train:
                loss.backward()
                optimizer.step()
                optimizer.zero_grad()

            np_z = z.cpu().data.numpy()
            np_x = x.cpu().data.numpy()
            np_y_idx = y_idx.cpu().data.numpy()

            metrics_epoch[f'{stage}_z'] += np_z.tolist()
            metrics_epoch[f'{stage}_y'] += np_y_idx.tolist()

        # calculate centers of the mass
        np_zs = np.array(metrics_epoch[f'{stage}_z'])
        np_ys = np.array(metrics_epoch[f'{stage}_y'])
        centers_by_classes = {}
        for y_idx in set(np_ys):
            matching_zs = np_zs[np_ys == y_idx]
            center = np.mean(matching_zs, axis=0)
            centers_by_classes[y_idx] = center

        y_prim_idx = []
        for z in np_zs:
            dists_to_centers = cdist(
                np.array(list(centers_by_classes.values())),
                np.expand_dims(z, axis=0),
                metric='euclidean'
            ).squeeze()
            idx_closest = np.argmin(dists_to_centers)
            y_prim_idx_each = list(centers_by_classes.keys())[idx_closest]
            y_prim_idx.append(y_prim_idx_each)

        np_y_prim_idx = np.array(y_prim_idx)
        acc = np.mean((np_y_prim_idx == np_ys) * 1.0)
        metrics_epoch[f'{stage}_acc'] = [acc]

    metrics_strs = []
    for key in metrics_epoch.keys():
        if '_z' not in key and '_y' not in key:
            value = 0
            if len(metrics_epoch[key]):
                value = np.mean(metrics_epoch[key])
            metrics[key].append(value)
            metrics_strs.append(f'{key}: {round(value, 2)}')
    print(f'epoch: {epoch} {" ".join(metrics_strs)}')

    plt.subplot(221) # row col idx
    plts = []
    c = 0
    for key, value in metrics.items():
        if '_z' in key or '_y' in key or '_dist' in key or 'test_' in key:
            continue
        plts += plt.plot(value, f'C{c}', label=key)
        ax = plt.twinx()
        c += 1

    plt.legend(plts, [it.get_label() for it in plts])

    for i, j in enumerate([4, 5, 6, 16, 17, 18, 10, 11, 12, 22, 23, 24]):
        plt.subplot(8, 6, j) # row col idx
        color = 'green' if np_y_idx[i] == np_y_prim_idx[i] else 'red'
        plt.title(f"y: {dataset_full.labels[np_y_idx[i]]} y_prim: {dataset_full.labels[np_y_prim_idx[i]]}", color=color)
        plt.imshow(np.transpose(np_x[i][0]), cmap='Greys')

    plt.subplot(223) # row col idx

    #pca = sklearn.manifold.TSNE(n_components=2)
    pca = sklearn.decomposition.PCA(n_components=2)

    MAX_POINTS = 100
    plt.title('train_z')
    np_z = np.array(metrics_epoch[f'train_z'])
    np_z = pca.fit_transform(np_z)
    np_y_idx = np.array(metrics_epoch[f'train_y'])
    labels = [dataset_full.labels[idx] for idx in np_y_idx[:MAX_POINTS]]
    set_labels = list(set(labels))
    c = [labels.index(it) for it in labels]
    scatter = plt.scatter(np_z[:MAX_POINTS, -1], np_z[:MAX_POINTS, -2], c=c, cmap=plt.get_cmap('prism'))
    plt.legend(
        handles=scatter.legend_elements()[0],
        loc='lower left',
        scatterpoints=1,
        fontsize=8,
        ncol=5,
        labels=set_labels
    )

    pca = sklearn.decomposition.PCA(n_components=2)

    plt.subplot(224) # row col idx

    plt.title('test_z')
    np_z = np.array(metrics_epoch[f'test_z'])
    np_z = pca.fit_transform(np_z)
    np_y_idx = np.array(metrics_epoch[f'test_y'])
    labels = [dataset_full.labels[idx] for idx in np_y_idx[:MAX_POINTS]]
    set_labels = list(set(labels))
    c = [labels.index(it) for it in labels]
    scatter = plt.scatter(np_z[:MAX_POINTS, -1], np_z[:MAX_POINTS, -2], c=c, cmap=plt.get_cmap('prism'))
    plt.legend(
        handles=scatter.legend_elements()[0],
        loc='lower left',
        scatterpoints=1,
        fontsize=8,
        ncol=5,
        labels=set_labels
    )

    plt.tight_layout(pad=0.5)


    if len(RUN_PATH) == 0:
        plt.show()
    else:
        if np.isnan(metrics[f'train_loss'][-1]) or np.isinf(metrics[f'test_loss'][-1]):
            exit()

        # save model weights
        plt.savefig(f'{RUN_PATH}/plt-{epoch}.png')
        #torch.save(model.state_dict(), f'{RUN_PATH}/model-{epoch}.pt')

